/* eslint-disable */
/**
 * Based on http://fileadmin.cs.lth.se/sde/publications/papers/2012-Soderberg-SCP-PrePrint.pdf
 */
let t;
import util from 'util';
function constructCfg(path, t, traverse) {
  let programEntry = createEntry();
  let programExit = createExit();

  return _construct(programEntry, node, programExit);

  return _construct(path.node);

  // CFGify - add CFG properties
  function cfgify(node) {
    if (!node.following) {
      node.following = new Set();
    }
    if (!node.succ) {
      node.succ = new Set();
    }
    if (!node.followingTrue) {
      node.followingTrue = null;
    }
    if (!node.followingFalse) {
      node.followingFalse = null;
    }
  }

  function isEntry(node) {
    return node.type === 'CFGEntry';
  }

  function isExit(node) {
    return node.type === 'CFGExit';
  }

  function createEntry() {
    return {
      type: 'CFGEntry',
      following: new Set()
    }
  }

  function createExit() {
    return {
      type: 'CFGExit'
    }
  }

  function follow(a, b) {
    cfgify(a);
    a.following.add(b);
  }

  function handleProgram(node) {
    let len = node.body.length;

    if (len === 0) {
      node.succ.add(programExit);
      return;
    }

    node.succ.add(node.body[0]);

    for (let i = 0; i < len; i++) {
      if (i === len - 1) {
        follow(node.body[i], programExit);
        _construct(node.body[i], programExit);
      } else {
        follow(node.body[i], node.body[i + 1]);
        _construct(node.body[i], node.body[i + 1]);
      }
    }
  }

  function handleBlockStatement(node, next) {
    let len = node.body.length;

    if (len === 0) {
      node.succ.add(next);
      return;
    }

    node.succ.add(node.body[0]);

    for (let i = 0; i < len; i++) {
      if (i === len - 1) {
        follow(node.body[i], next);
        _construct(node.body[i], next);
      } else {
        follow(node.body[i], node.body[i + 1]);
        _construct(node.body[i], node.body[i + 1]);
      }
    }
  }

  function handleVariableDeclaration(node, next) {
    node.succ = node.following;
  }

  function handleIfStatement(node, next) {
    node.succ.add(node.test);
    cfgify(node.test);

    node.test.followingTrue = node.consequent;
    node.test.followingFalse = node.alternate ? node.alternate : next;
    node.test.following.add(node.test.followingTrue, node.test.followingFalse);

    _construct(node.test, node.test.followingTrue);
  }

  function _construct(entry, node, exit) {
    if (node.cfgVisited) {
      console.log('loop at', node.type);
      return node;
    }

    cfgify(node);

    node.cfgVisited = true;

    if (t.isProgram(node)) {
      return handleProgram(entry, node, exit);
    }

    if (t.isBlockStatement(node)) {
      return handleBlockStatement(entry, node);
    }

    if (t.isVariableDeclaration(node)) {
      return handleVariableDeclaration(entry, node);
    }

    if (t.isIfStatement(node)) {
      return handleIfStatement(entry, node);
    }
  }
}

export default ({types, traverse}) => ({
  visitor: {
    Program(path) {
      t = types;
      constructCfg(path, types, traverse);
      const {node} = path;
      print(node.entry);
    }
  }
});

function print(entry) {
  let visited = new Set();

  padlog(0, entry.type);
  _print(entry);

  function padlog(n, str) {
    console.log('  '.repeat(n) + str);
  }
  function _print(node, n = 0) {
    if (visited.has(node)) {
      console.log('loop at', node.type);
      return;
    }
    visited.add(node);

    if (node.succ) {
      for (let i of node.succ) {
        padlog(n, i.type);
        if (t.isBlockStatement(i)) {
          padlog(n + 1, i.entry.type);
          _print(i.entry, n + 1);
        }
        if (t.isIfStatement(i)) {
          padlog(n + 1, i.followingTrue.type);
          _print(i.followingTrue, n + 1);
          if (i.followingFalse) {
            padlog(n + 1, i.followingFalse.type);
            _print(i.followingFalse, n + 1);
          }
        }
        _print(i, n);
      }
    }
  }
}
